/* This is a generated file, don't edit */

#define NUM_APPLETS 195
#define KNOWN_APPNAME_OFFSETS 8

const uint16_t applet_nameofs[] ALIGN2 = {
137,
293,
443,
594,
758,
919,
1077,
};

const char applet_names[] ALIGN1 = ""
"[" "\0"
"[[" "\0"
"addgroup" "\0"
"adduser" "\0"
"ascii" "\0"
"ash" "\0"
"awk" "\0"
"base32" "\0"
"basename" "\0"
"blkid" "\0"
"bunzip2" "\0"
"bzcat" "\0"
"bzip2" "\0"
"cat" "\0"
"chattr" "\0"
"chgrp" "\0"
"chmod" "\0"
"chown" "\0"
"chroot" "\0"
"chvt" "\0"
"clear" "\0"
"cmp" "\0"
"cp" "\0"
"cpio" "\0"
"crc32" "\0"
"cttyhack" "\0"
"cut" "\0"
"date" "\0"
"dc" "\0"
"dd" "\0"
"deallocvt" "\0"
"delgroup" "\0"
"deluser" "\0"
"depmod" "\0"
"devmem" "\0"
"df" "\0"
"diff" "\0"
"dirname" "\0"
"dmesg" "\0"
"dnsdomainname" "\0"
"du" "\0"
"dumpkmap" "\0"
"dumpleases" "\0"
"echo" "\0"
"egrep" "\0"
"env" "\0"
"expr" "\0"
"false" "\0"
"fbset" "\0"
"fdisk" "\0"
"fgrep" "\0"
"find" "\0"
"flock" "\0"
"free" "\0"
"fsck" "\0"
"fstrim" "\0"
"ftpd" "\0"
"ftpget" "\0"
"ftpput" "\0"
"fuser" "\0"
"getopt" "\0"
"getty" "\0"
"grep" "\0"
"groups" "\0"
"gunzip" "\0"
"gzip" "\0"
"hd" "\0"
"head" "\0"
"hexdump" "\0"
"hostname" "\0"
"httpd" "\0"
"hwclock" "\0"
"id" "\0"
"ifconfig" "\0"
"ifdown" "\0"
"ifup" "\0"
"inetd" "\0"
"insmod" "\0"
"ip" "\0"
"kill" "\0"
"killall" "\0"
"klogd" "\0"
"less" "\0"
"ln" "\0"
"loadfont" "\0"
"loadkmap" "\0"
"logger" "\0"
"logname" "\0"
"logread" "\0"
"losetup" "\0"
"ls" "\0"
"lsmod" "\0"
"lspci" "\0"
"lzcat" "\0"
"md5sum" "\0"
"mdev" "\0"
"mesg" "\0"
"microcom" "\0"
"mkdir" "\0"
"mkfifo" "\0"
"mkfs.vfat" "\0"
"mknod" "\0"
"mkswap" "\0"
"mktemp" "\0"
"modprobe" "\0"
"more" "\0"
"mount" "\0"
"mountpoint" "\0"
"mv" "\0"
"nc" "\0"
"netstat" "\0"
"nohup" "\0"
"nproc" "\0"
"nslookup" "\0"
"od" "\0"
"openvt" "\0"
"patch" "\0"
"pgrep" "\0"
"pidof" "\0"
"pivot_root" "\0"
"printf" "\0"
"ps" "\0"
"pwd" "\0"
"rdate" "\0"
"readlink" "\0"
"realpath" "\0"
"reboot" "\0"
"renice" "\0"
"reset" "\0"
"resize" "\0"
"rev" "\0"
"rm" "\0"
"rmdir" "\0"
"rmmod" "\0"
"route" "\0"
"run-parts" "\0"
"sed" "\0"
"seq" "\0"
"setconsole" "\0"
"setsid" "\0"
"sh" "\0"
"sha1sum" "\0"
"sha256sum" "\0"
"sha512sum" "\0"
"shuf" "\0"
"sleep" "\0"
"sort" "\0"
"start-stop-daemon" "\0"
"stat" "\0"
"strings" "\0"
"stty" "\0"
"sulogin" "\0"
"swapoff" "\0"
"swapon" "\0"
"switch_root" "\0"
"sync" "\0"
"sysctl" "\0"
"syslogd" "\0"
"tail" "\0"
"tar" "\0"
"tee" "\0"
"telnet" "\0"
"telnetd" "\0"
"test" "\0"
"tftp" "\0"
"tftpd" "\0"
"time" "\0"
"top" "\0"
"touch" "\0"
"tr" "\0"
"true" "\0"
"ts" "\0"
"tty" "\0"
"udhcpc" "\0"
"udhcpd" "\0"
"umount" "\0"
"uname" "\0"
"uniq" "\0"
"unlink" "\0"
"unzip" "\0"
"uptime" "\0"
"users" "\0"
"usleep" "\0"
"vi" "\0"
"watch" "\0"
"watchdog" "\0"
"wc" "\0"
"wget" "\0"
"which" "\0"
"who" "\0"
"whoami" "\0"
"xargs" "\0"
"xzcat" "\0"
"yes" "\0"
"zcat" "\0"
;

#define APPLET_NO_addgroup 2
#define APPLET_NO_adduser 3
#define APPLET_NO_ascii 4
#define APPLET_NO_ash 5
#define APPLET_NO_awk 6
#define APPLET_NO_base32 7
#define APPLET_NO_basename 8
#define APPLET_NO_blkid 9
#define APPLET_NO_bunzip2 10
#define APPLET_NO_bzcat 11
#define APPLET_NO_bzip2 12
#define APPLET_NO_cat 13
#define APPLET_NO_chattr 14
#define APPLET_NO_chgrp 15
#define APPLET_NO_chmod 16
#define APPLET_NO_chown 17
#define APPLET_NO_chroot 18
#define APPLET_NO_chvt 19
#define APPLET_NO_clear 20
#define APPLET_NO_cmp 21
#define APPLET_NO_cp 22
#define APPLET_NO_cpio 23
#define APPLET_NO_crc32 24
#define APPLET_NO_cttyhack 25
#define APPLET_NO_cut 26
#define APPLET_NO_date 27
#define APPLET_NO_dc 28
#define APPLET_NO_dd 29
#define APPLET_NO_deallocvt 30
#define APPLET_NO_delgroup 31
#define APPLET_NO_deluser 32
#define APPLET_NO_depmod 33
#define APPLET_NO_devmem 34
#define APPLET_NO_df 35
#define APPLET_NO_diff 36
#define APPLET_NO_dirname 37
#define APPLET_NO_dmesg 38
#define APPLET_NO_dnsdomainname 39
#define APPLET_NO_du 40
#define APPLET_NO_dumpkmap 41
#define APPLET_NO_dumpleases 42
#define APPLET_NO_echo 43
#define APPLET_NO_egrep 44
#define APPLET_NO_env 45
#define APPLET_NO_expr 46
#define APPLET_NO_false 47
#define APPLET_NO_fbset 48
#define APPLET_NO_fdisk 49
#define APPLET_NO_fgrep 50
#define APPLET_NO_find 51
#define APPLET_NO_flock 52
#define APPLET_NO_free 53
#define APPLET_NO_fsck 54
#define APPLET_NO_fstrim 55
#define APPLET_NO_ftpd 56
#define APPLET_NO_ftpget 57
#define APPLET_NO_ftpput 58
#define APPLET_NO_fuser 59
#define APPLET_NO_getopt 60
#define APPLET_NO_getty 61
#define APPLET_NO_grep 62
#define APPLET_NO_groups 63
#define APPLET_NO_gunzip 64
#define APPLET_NO_gzip 65
#define APPLET_NO_hd 66
#define APPLET_NO_head 67
#define APPLET_NO_hexdump 68
#define APPLET_NO_hostname 69
#define APPLET_NO_httpd 70
#define APPLET_NO_hwclock 71
#define APPLET_NO_id 72
#define APPLET_NO_ifconfig 73
#define APPLET_NO_ifdown 74
#define APPLET_NO_ifup 75
#define APPLET_NO_inetd 76
#define APPLET_NO_insmod 77
#define APPLET_NO_ip 78
#define APPLET_NO_kill 79
#define APPLET_NO_killall 80
#define APPLET_NO_klogd 81
#define APPLET_NO_less 82
#define APPLET_NO_ln 83
#define APPLET_NO_loadfont 84
#define APPLET_NO_loadkmap 85
#define APPLET_NO_logger 86
#define APPLET_NO_logname 87
#define APPLET_NO_logread 88
#define APPLET_NO_losetup 89
#define APPLET_NO_ls 90
#define APPLET_NO_lsmod 91
#define APPLET_NO_lspci 92
#define APPLET_NO_lzcat 93
#define APPLET_NO_md5sum 94
#define APPLET_NO_mdev 95
#define APPLET_NO_mesg 96
#define APPLET_NO_microcom 97
#define APPLET_NO_mkdir 98
#define APPLET_NO_mkfifo 99
#define APPLET_NO_mknod 101
#define APPLET_NO_mkswap 102
#define APPLET_NO_mktemp 103
#define APPLET_NO_modprobe 104
#define APPLET_NO_more 105
#define APPLET_NO_mount 106
#define APPLET_NO_mountpoint 107
#define APPLET_NO_mv 108
#define APPLET_NO_nc 109
#define APPLET_NO_netstat 110
#define APPLET_NO_nohup 111
#define APPLET_NO_nproc 112
#define APPLET_NO_nslookup 113
#define APPLET_NO_od 114
#define APPLET_NO_openvt 115
#define APPLET_NO_patch 116
#define APPLET_NO_pgrep 117
#define APPLET_NO_pidof 118
#define APPLET_NO_pivot_root 119
#define APPLET_NO_printf 120
#define APPLET_NO_ps 121
#define APPLET_NO_pwd 122
#define APPLET_NO_rdate 123
#define APPLET_NO_readlink 124
#define APPLET_NO_realpath 125
#define APPLET_NO_reboot 126
#define APPLET_NO_renice 127
#define APPLET_NO_reset 128
#define APPLET_NO_resize 129
#define APPLET_NO_rev 130
#define APPLET_NO_rm 131
#define APPLET_NO_rmdir 132
#define APPLET_NO_rmmod 133
#define APPLET_NO_route 134
#define APPLET_NO_sed 136
#define APPLET_NO_seq 137
#define APPLET_NO_setconsole 138
#define APPLET_NO_setsid 139
#define APPLET_NO_sh 140
#define APPLET_NO_sha1sum 141
#define APPLET_NO_sha256sum 142
#define APPLET_NO_sha512sum 143
#define APPLET_NO_shuf 144
#define APPLET_NO_sleep 145
#define APPLET_NO_sort 146
#define APPLET_NO_stat 148
#define APPLET_NO_strings 149
#define APPLET_NO_stty 150
#define APPLET_NO_sulogin 151
#define APPLET_NO_swapoff 152
#define APPLET_NO_swapon 153
#define APPLET_NO_switch_root 154
#define APPLET_NO_sync 155
#define APPLET_NO_sysctl 156
#define APPLET_NO_syslogd 157
#define APPLET_NO_tail 158
#define APPLET_NO_tar 159
#define APPLET_NO_tee 160
#define APPLET_NO_telnet 161
#define APPLET_NO_telnetd 162
#define APPLET_NO_test 163
#define APPLET_NO_tftp 164
#define APPLET_NO_tftpd 165
#define APPLET_NO_time 166
#define APPLET_NO_top 167
#define APPLET_NO_touch 168
#define APPLET_NO_tr 169
#define APPLET_NO_true 170
#define APPLET_NO_ts 171
#define APPLET_NO_tty 172
#define APPLET_NO_udhcpc 173
#define APPLET_NO_udhcpd 174
#define APPLET_NO_umount 175
#define APPLET_NO_uname 176
#define APPLET_NO_uniq 177
#define APPLET_NO_unlink 178
#define APPLET_NO_unzip 179
#define APPLET_NO_uptime 180
#define APPLET_NO_users 181
#define APPLET_NO_usleep 182
#define APPLET_NO_vi 183
#define APPLET_NO_watch 184
#define APPLET_NO_watchdog 185
#define APPLET_NO_wc 186
#define APPLET_NO_wget 187
#define APPLET_NO_which 188
#define APPLET_NO_who 189
#define APPLET_NO_whoami 190
#define APPLET_NO_xargs 191
#define APPLET_NO_xzcat 192
#define APPLET_NO_yes 193
#define APPLET_NO_zcat 194

#ifndef SKIP_applet_main
int (*const applet_main[])(int argc, char **argv) = {
test_main,
test_main,
addgroup_main,
adduser_main,
ascii_main,
ash_main,
awk_main,
baseNUM_main,
basename_main,
blkid_main,
bunzip2_main,
bunzip2_main,
bzip2_main,
cat_main,
chattr_main,
chgrp_main,
chmod_main,
chown_main,
chroot_main,
chvt_main,
clear_main,
cmp_main,
cp_main,
cpio_main,
cksum_main,
cttyhack_main,
cut_main,
date_main,
dc_main,
dd_main,
deallocvt_main,
deluser_main,
deluser_main,
depmod_main,
devmem_main,
df_main,
diff_main,
dirname_main,
dmesg_main,
hostname_main,
du_main,
dumpkmap_main,
dumpleases_main,
echo_main,
grep_main,
env_main,
expr_main,
false_main,
fbset_main,
fdisk_main,
grep_main,
find_main,
flock_main,
free_main,
fsck_main,
fstrim_main,
ftpd_main,
ftpgetput_main,
ftpgetput_main,
fuser_main,
getopt_main,
getty_main,
grep_main,
id_main,
gunzip_main,
gzip_main,
hexdump_main,
head_main,
hexdump_main,
hostname_main,
httpd_main,
hwclock_main,
id_main,
ifconfig_main,
ifupdown_main,
ifupdown_main,
inetd_main,
insmod_main,
ip_main,
kill_main,
kill_main,
klogd_main,
less_main,
ln_main,
loadfont_main,
loadkmap_main,
logger_main,
logname_main,
logread_main,
losetup_main,
ls_main,
lsmod_main,
lspci_main,
unlzma_main,
md5_sha1_sum_main,
mdev_main,
mesg_main,
microcom_main,
mkdir_main,
mkfifo_main,
mkfs_vfat_main,
mknod_main,
mkswap_main,
mktemp_main,
modprobe_main,
more_main,
mount_main,
mountpoint_main,
mv_main,
nc_main,
netstat_main,
nohup_main,
nproc_main,
nslookup_main,
od_main,
openvt_main,
patch_main,
pgrep_main,
pidof_main,
pivot_root_main,
printf_main,
ps_main,
pwd_main,
rdate_main,
readlink_main,
realpath_main,
halt_main,
renice_main,
reset_main,
resize_main,
rev_main,
rm_main,
rmdir_main,
rmmod_main,
route_main,
run_parts_main,
sed_main,
seq_main,
setconsole_main,
setsid_main,
ash_main,
md5_sha1_sum_main,
md5_sha1_sum_main,
md5_sha1_sum_main,
shuf_main,
sleep_main,
sort_main,
start_stop_daemon_main,
stat_main,
strings_main,
stty_main,
sulogin_main,
swap_on_off_main,
swap_on_off_main,
switch_root_main,
sync_main,
sysctl_main,
syslogd_main,
tail_main,
tar_main,
tee_main,
telnet_main,
telnetd_main,
test_main,
tftp_main,
tftpd_main,
time_main,
top_main,
touch_main,
tr_main,
true_main,
ts_main,
tty_main,
udhcpc_main,
udhcpd_main,
umount_main,
uname_main,
uniq_main,
unlink_main,
unzip_main,
uptime_main,
who_main,
usleep_main,
vi_main,
watch_main,
watchdog_main,
wc_main,
wget_main,
which_main,
who_main,
whoami_main,
xargs_main,
unxz_main,
yes_main,
gunzip_main,
};
#endif

const uint8_t applet_suid[] ALIGN1 = {
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
0x00,
};

